@@ -23,15 +23,20 @@
 #ifndef GEANY_FILETYPES_H
 #define GEANY_FILETYPES_H 1
 
+#include "geany.h" /* for GEANY() macro */
+#include "tm_source_file.h" /* for langType */
+
+#include "gtkcompat.h" /* Needed by ScintillaWidget.h */
+#include "Scintilla.h" /* Needed by ScintillaWidget.h */
+#include "ScintillaWidget.h" /* for ScintillaObject */
+
+#include <glib.h>
+
+
 G_BEGIN_DECLS
 
-#include "Scintilla.h"
-#include "ScintillaWidget.h"
-
-#ifdef GEANY_PRIVATE
-#include "build.h"
-#endif
-
+/* Forward-declared to avoid including document.h since it includes this header */
+struct GeanyDocument;
 
 /* Do not change the order, only append. */
 typedef enum
@@ -95,6 +100,9 @@ typedef enum
 	GEANY_FILETYPES_BATCH,
 	GEANY_FILETYPES_POWERSHELL,
 	GEANY_FILETYPES_RUST,
+	GEANY_FILETYPES_COFFEESCRIPT,
+	GEANY_FILETYPES_GO,
+	GEANY_FILETYPES_ZEPHIR,
 	/* ^ append items here */
 	GEANY_MAX_BUILT_IN_FILETYPES	/* Don't use this, use filetypes_array->len instead */
 }
@@ -118,7 +126,7 @@ GeanyFiletypeGroupID;
 	(((filetype_ptr) != NULL) ? (filetype_ptr)->id : GEANY_FILETYPES_NONE)
 
 /** Represents a filetype. */
-struct GeanyFiletype
+typedef struct GeanyFiletype
 {
 	filetype_id		  id;				/**< Index in @c filetypes_array. */
 	/** Represents the langType of tagmanager (see the table
@@ -138,30 +146,17 @@ struct GeanyFiletype
 	gboolean		  comment_use_indent;
 	GeanyFiletypeGroupID group;
 	gchar			 *error_regex_string;
-	GeanyFiletype	 *lexer_filetype;
+	struct GeanyFiletype	 *lexer_filetype;
 	gchar			 *mime_type;
-	GdkPixbuf		 *icon;
+	GIcon			 *icon;
 	gchar			 *comment_single; /* single-line comment */
 	/* filetype indent settings, -1 if not set */
 	gint			  indent_type;
 	gint			  indent_width;
 
 	struct GeanyFiletypePrivate	*priv;	/* must be last, append fields before this item */
-#ifdef GEANY_PRIVATE
-	/* Do not use following fields in plugins */
-	/* TODO: move these fields into filetypesprivate.h */
-	GeanyBuildCommand *filecmds;
-	GeanyBuildCommand *ftdefcmds;
-	GeanyBuildCommand *execcmds;
-	GeanyBuildCommand *homefilecmds;
-	GeanyBuildCommand *homeexeccmds;
-	GeanyBuildCommand *projfilecmds;
-	GeanyBuildCommand *projexeccmds;
-	gint			 project_list_entry;
-	gchar			 *projerror_regex_string;
-	gchar			 *homeerror_regex_string;
-#endif
-};
+}
+GeanyFiletype;
 
 extern GPtrArray *filetypes_array;
 
@@ -170,11 +165,23 @@ extern GPtrArray *filetypes_array;
  * @see filetypes_index(). */
 #define filetypes	((GeanyFiletype **)GEANY(filetypes_array)->pdata)
 
-extern GSList *filetypes_by_title;
 
+GeanyFiletype *filetypes_detect_from_file(const gchar *utf8_filename);
 
 GeanyFiletype *filetypes_lookup_by_name(const gchar *name);
 
+GeanyFiletype *filetypes_index(gint idx);
+
+const gchar *filetypes_get_display_name(GeanyFiletype *ft);
+
+const GSList *filetypes_get_sorted_by_name(void);
+
+
+#ifdef GEANY_PRIVATE
+
+extern GSList *filetypes_by_title;
+
+
 GeanyFiletype *filetypes_find(GCompareFunc predicate, gpointer user_data);
 
 
@@ -187,18 +194,10 @@ void filetypes_reload_extensions(void);
 void filetypes_reload(void);
 
 
-GeanyFiletype *filetypes_index(gint idx);
-
-const GSList *filetypes_get_sorted_by_name(void);
-
-const gchar *filetypes_get_display_name(GeanyFiletype *ft);
-
-GeanyFiletype *filetypes_detect_from_document(GeanyDocument *doc);
+GeanyFiletype *filetypes_detect_from_document(struct GeanyDocument *doc);
 
 GeanyFiletype *filetypes_detect_from_extension(const gchar *utf8_filename);
 
-GeanyFiletype *filetypes_detect_from_file(const gchar *utf8_filename);
-
 void filetypes_free_types(void);
 
 void filetypes_load_config(guint ft_id, gboolean reload);
@@ -219,6 +218,8 @@ gboolean filetypes_parse_error_message(GeanyFiletype *ft, const gchar *message,
 gboolean filetype_get_comment_open_close(const GeanyFiletype *ft, gboolean single_first,
 		const gchar **co, const gchar **cc);
 
+#endif /* GEANY_PRIVATE */
+
 G_END_DECLS
 
-#endif
+#endif /* GEANY_FILETYPES_H */
