@@ -28,38 +28,42 @@
  * filename_xx=pos;filetype UID;read only;Eencoding;use_tabs;auto_indent;line_wrapping;filename
  */
 
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include "keyfile.h"
+
+#include "app.h"
+#include "build.h"
+#include "document.h"
+#include "encodings.h"
+#include "filetypes.h"
+#include "geanyobject.h"
+#include "main.h"
+#include "msgwindow.h"
+#include "prefs.h"
+#include "printing.h"
+#include "project.h"
+#include "sciwrappers.h"
+#include "stash.h"
+#include "support.h"
+#include "templates.h"
+#include "toolbar.h"
+#include "ui_utils.h"
+#include "utils.h"
+#include "vte.h"
+
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
 
-#include "geany.h"
-
 #ifdef HAVE_VTE
 #include <pwd.h>
 #include <sys/types.h>
 #include <unistd.h>
 #endif
 
-#include "support.h"
-#include "keyfile.h"
-#include "prefs.h"
-#include "ui_utils.h"
-#include "utils.h"
-#include "document.h"
-#include "filetypes.h"
-#include "sciwrappers.h"
-#include "encodings.h"
-#include "vte.h"
-#include "main.h"
-#include "msgwindow.h"
-#include "search.h"
-#include "project.h"
-#include "editor.h"
-#include "printing.h"
-#include "templates.h"
-#include "toolbar.h"
-#include "stash.h"
-
 
 /* some default settings which are used at the very first start of Geany to fill
  * the configuration file */
@@ -70,16 +74,25 @@
 #define GEANY_DEFAULT_TOOLS_MAKE		"make"
 #ifdef G_OS_WIN32
 #define GEANY_DEFAULT_TOOLS_TERMINAL	"cmd.exe /Q /C %c"
+#elif defined(__APPLE__)
+#define GEANY_DEFAULT_TOOLS_TERMINAL	"open -a terminal %c"
 #else
 #define GEANY_DEFAULT_TOOLS_TERMINAL	"xterm -e \"/bin/sh %c\""
 #endif
+#ifdef __APPLE__
+#define GEANY_DEFAULT_TOOLS_BROWSER		"open -a safari"
+#define GEANY_DEFAULT_FONT_SYMBOL_LIST	"Helvetica Medium 12"
+#define GEANY_DEFAULT_FONT_MSG_WINDOW	"Menlo Medium 12"
+#define GEANY_DEFAULT_FONT_EDITOR		"Menlo Medium 12"
+#else
 #define GEANY_DEFAULT_TOOLS_BROWSER		"firefox"
+#define GEANY_DEFAULT_FONT_SYMBOL_LIST	"Sans 9"
+#define GEANY_DEFAULT_FONT_MSG_WINDOW	"Monospace 9"
+#define GEANY_DEFAULT_FONT_EDITOR		"Monospace 10"
+#endif
 #define GEANY_DEFAULT_TOOLS_PRINTCMD	"lpr"
 #define GEANY_DEFAULT_TOOLS_GREP		"grep"
 #define GEANY_DEFAULT_MRU_LENGTH		10
-#define GEANY_DEFAULT_FONT_SYMBOL_LIST	"Sans 9"
-#define GEANY_DEFAULT_FONT_MSG_WINDOW	"Sans 9"
-#define GEANY_DEFAULT_FONT_EDITOR		"Monospace 10"
 #define GEANY_TOGGLE_MARK				"~ "
 #define GEANY_MAX_AUTOCOMPLETE_WORDS	30
 #define GEANY_MAX_SYMBOLS_UPDATE_FREQ	250
@@ -220,6 +233,10 @@ static void init_pref_groups(void)
 		"gio_unsafe_save_backup", FALSE);
 	stash_group_add_boolean(group, &file_prefs.use_gio_unsafe_file_saving,
 		"use_gio_unsafe_file_saving", TRUE);
+	stash_group_add_boolean(group, &file_prefs.keep_edit_history_on_reload,
+		"keep_edit_history_on_reload", TRUE);
+	stash_group_add_boolean(group, &file_prefs.show_keep_edit_history_on_reload_msg,
+		"show_keep_edit_history_on_reload_msg", TRUE);
 	/* for backwards-compatibility */
 	stash_group_add_integer(group, &editor_prefs.indentation->hard_tab_width,
 		"indent_hard_tab_width", 8);
@@ -227,6 +244,8 @@ static void init_pref_groups(void)
 		"find_selection_type", GEANY_FIND_SEL_CURRENT_WORD);
 	stash_group_add_string(group, &file_prefs.extract_filetype_regex,
 		"extract_filetype_regex", GEANY_DEFAULT_FILETYPE_REGEX);
+	stash_group_add_boolean(group, &search_prefs.replace_and_find_by_default,
+		"replace_and_find_by_default", TRUE);
 
 	/* Note: Interface-related various prefs are in ui_init_prefs() */
 
@@ -325,10 +344,23 @@ static gchar *get_session_file_string(GeanyDocument *doc)
 }
 
 
+static void remove_session_files(GKeyFile *config)
+{
+	gchar **ptr;
+	gchar **keys = g_key_file_get_keys(config, "files", NULL, NULL);
+
+	foreach_strv(ptr, keys)
+	{
+		if (g_str_has_prefix(*ptr, "FILE_NAME_"))
+			g_key_file_remove_key(config, "files", *ptr, NULL);
+	}
+	g_strfreev(keys);
+}
+
+
 void configuration_save_session_files(GKeyFile *config)
 {
 	gint npage;
-	gchar *tmp;
 	gchar entry[16];
 	guint i = 0, j = 0, max;
 	GeanyDocument *doc;
@@ -336,6 +368,9 @@ void configuration_save_session_files(GKeyFile *config)
 	npage = gtk_notebook_get_current_page(GTK_NOTEBOOK(main_widgets.notebook));
 	g_key_file_set_integer(config, "files", "current_page", npage);
 
+	// clear existing entries first as they might not all be overwritten
+	remove_session_files(config);
+
 	/* store the filenames in the notebook tab order to reopen them the next time */
 	max = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook));
 	for (i = 0; i < max; i++)
@@ -352,23 +387,6 @@ void configuration_save_session_files(GKeyFile *config)
 			j++;
 		}
 	}
-	/* if open filenames less than saved session files, delete existing entries in the list */
-	i = j;
-	while (TRUE)
-	{
-		g_snprintf(entry, sizeof(entry), "FILE_NAME_%d", i);
-		tmp = g_key_file_get_string(config, "files", entry, NULL);
-		if (G_UNLIKELY(tmp == NULL))
-		{
-			break;
-		}
-		else
-		{
-			g_key_file_remove_key(config, "files", entry, NULL);
-			g_free(tmp);
-			i++;
-		}
-	}
 
 #ifdef HAVE_VTE
 	if (vte_info.have_vte)
@@ -657,7 +675,11 @@ void configuration_load_session_files(GKeyFile *config, gboolean read_recent_fil
 
 	/* the project may load another list than the main setting */
 	if (session_files != NULL)
+	{
+		foreach_ptr_array(tmp_array, i, session_files)
+			g_strfreev(tmp_array);
 		g_ptr_array_free(session_files, TRUE);
+	}
 
 	session_files = g_ptr_array_new();
 	have_session_files = TRUE;
@@ -841,6 +863,13 @@ static void load_dialog_prefs(GKeyFile *config)
 		struct passwd *pw = getpwuid(getuid());
 		const gchar *shell = (pw != NULL) ? pw->pw_shell : "/bin/sh";
 
+#ifdef __APPLE__
+		/* Geany is started using launchd on OS X and we don't get any environment variables
+		 * so PS1 isn't defined. Start as a login shell to read the corresponding config files. */
+		if (strcmp(shell, "/bin/bash") == 0)
+			shell = "/bin/bash -l";
+#endif
+
 		vc = g_new0(VteConfig, 1);
 		vte_info.dir = utils_get_setting_string(config, "VTE", "last_dir", NULL);
 		if ((vte_info.dir == NULL || utils_str_equal(vte_info.dir, "")) && pw != NULL)
@@ -855,7 +884,7 @@ static void load_dialog_prefs(GKeyFile *config)
 			"send_selection_unsafe", FALSE);
 		vc->image = utils_get_setting_string(config, "VTE", "image", "");
 		vc->shell = utils_get_setting_string(config, "VTE", "shell", shell);
-		vc->font = utils_get_setting_string(config, "VTE", "font", "Monospace 10");
+		vc->font = utils_get_setting_string(config, "VTE", "font", GEANY_DEFAULT_FONT_EDITOR);
 		vc->scroll_on_key = utils_get_setting_boolean(config, "VTE", "scroll_on_key", TRUE);
 		vc->scroll_on_out = utils_get_setting_boolean(config, "VTE", "scroll_on_out", TRUE);
 		vc->enable_bash_keys = utils_get_setting_boolean(config, "VTE", "enable_bash_keys", TRUE);
@@ -921,19 +950,20 @@ static void load_dialog_prefs(GKeyFile *config)
 
 	/* printing */
 	tmp_string2 = g_find_program_in_path(GEANY_DEFAULT_TOOLS_PRINTCMD);
-#ifdef G_OS_WIN32
+
 	if (!EMPTY(tmp_string2))
 	{
-		/* single quote paths on Win32 for g_spawn_command_line_async */
-		tmp_string = g_strconcat("'", tmp_string2, "' '%f'", NULL);
+	#ifdef G_OS_WIN32
+		tmp_string = g_strconcat(GEANY_DEFAULT_TOOLS_PRINTCMD, " \"%f\"", NULL);
+	#else
+		tmp_string = g_strconcat(GEANY_DEFAULT_TOOLS_PRINTCMD, " '%f'", NULL);
+	#endif
 	}
 	else
 	{
 		tmp_string = g_strdup("");
 	}
-#else
-	tmp_string = g_strconcat(tmp_string2, " %f", NULL);
-#endif
+
 	printing_prefs.external_print_cmd = utils_get_setting_string(config, "printing", "print_cmd", tmp_string);
 	g_free(tmp_string);
 	g_free(tmp_string2);
@@ -1043,6 +1073,27 @@ void configuration_save_default_session(void)
 }
 
 
+void configuration_clear_default_session(void)
+{
+	gchar *configfile = g_build_filename(app->configdir, "geany.conf", NULL);
+	gchar *data;
+	GKeyFile *config = g_key_file_new();
+
+	g_key_file_load_from_file(config, configfile, G_KEY_FILE_NONE, NULL);
+
+	if (cl_options.load_session)
+		remove_session_files(config);
+
+	/* write the file */
+	data = g_key_file_to_data(config, NULL, NULL);
+	utils_write_file(configfile, data);
+	g_free(data);
+
+	g_key_file_free(config);
+	g_free(configfile);
+}
+
+
 /*
  * Only reload the session part of the default configuration
  */
@@ -1194,14 +1245,20 @@ void configuration_open_files(void)
 
 	if (failure)
 		ui_set_statusbar(TRUE, _("Failed to load one or more session files."));
-	else if (session_notebook_page >= 0)
+	else
 	{
-		/* explicitly allow notebook switch page callback to be called for window title,
-		 * encoding settings and so other things */
+		/* explicitly trigger a notebook page switch after unsetting main_status.opening_session_files
+		 * for callbacks to run (and update window title, encoding settings, and so on) */
+		gint n_pages = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook));
+		gint cur_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(main_widgets.notebook));
+		gint target_page = session_notebook_page >= 0 ? session_notebook_page : cur_page;
+
+		/* if target page is current page, switch to another page first to really trigger an event */
+		if (target_page == cur_page && n_pages > 0)
+			gtk_notebook_set_current_page(GTK_NOTEBOOK(main_widgets.notebook), (cur_page + 1) % n_pages);
+
 		main_status.opening_session_files = FALSE;
-		/** TODO if session_notebook_page is equal to the current notebook tab(the last opened)
-		 ** the notebook switch page callback isn't triggered and e.g. menu items are not updated */
-		gtk_notebook_set_current_page(GTK_NOTEBOOK(main_widgets.notebook), session_notebook_page);
+		gtk_notebook_set_current_page(GTK_NOTEBOOK(main_widgets.notebook), target_page);
 	}
 	main_status.opening_session_files = FALSE;
 }
