@@ -1,8 +1,9 @@
 /*
  *      socket.c - this file is part of Geany, a fast and lightweight IDE
  *
- *      Copyright 2006-2008 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
- *      Copyright 2006-2008 Nick Treleaven <nick(dot)treleaven(at)btinternet(dot)com>
+ *      Copyright 2006-2012 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
+ *      Copyright 2006-2012 Nick Treleaven <nick(dot)treleaven(at)btinternet(dot)com>
+ *      Copyright 2006 Hiroyuki Yamamoto (author of Sylpheed)
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -14,11 +15,9 @@
  *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *      GNU General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- *
- * $Id: socket.c 2991 2008-09-24 12:07:22Z ntrel $
+ *      You should have received a copy of the GNU General Public License along
+ *      with this program; if not, write to the Free Software Foundation, Inc.,
+ *      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 /*
@@ -38,8 +37,10 @@
  * .\n
  * The first thing should be the command name followed by the data belonging to the command and
  * to mark the end of data send a single '.'. Each message should be ended with \n.
+ * The command window is only available on Windows and takes no additional data, instead it
+ * writes back a Windows handle (HWND) for the main window to set it to the foreground (focus).
  *
- * At the moment the commands open, line and column are available.
+ * At the moment the commands window, doclist, open, openro, line and column are available.
  *
  * About the socket files on Unix-like systems:
  * Geany creates a socket in /tmp (or any other directory returned by g_get_tmp_dir()) and
@@ -50,13 +51,27 @@
  *
  */
 
-
-#include "geany.h"
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
 
 #ifdef HAVE_SOCKET
 
+#include "socket.h"
+
+#include "app.h"
+#include "dialogs.h"
+#include "document.h"
+#include "encodings.h"
+#include "main.h"
+#include "support.h"
+#include "utils.h"
+#include "win32.h"
+
+#include "gtkcompat.h"
+
+
 #ifndef G_OS_WIN32
-# include <string.h>
 # include <sys/time.h>
 # include <sys/types.h>
 # include <sys/socket.h>
@@ -65,32 +80,28 @@
 # include <glib/gstdio.h>
 #else
 # include <winsock2.h>
+# include <windows.h>
+# include <gdk/gdkwin32.h>
 # include <ws2tcpip.h>
 #endif
+#include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
 
-#include "main.h"
-#include "socket.h"
-#include "document.h"
-#include "support.h"
-#include "ui_utils.h"
-#include "utils.h"
-#include "encodings.h"
-
+#ifdef GDK_WINDOWING_X11
+#include <gdk/gdkx.h>
+#endif
 
 
 #ifdef G_OS_WIN32
 #define REMOTE_CMD_PORT		49876
-#define SockDesc			SOCKET
 #define SOCKET_IS_VALID(s)	((s) != INVALID_SOCKET)
 #else
-#define SockDesc			gint
 #define SOCKET_IS_VALID(s)	((s) >= 0)
 #define INVALID_SOCKET		(-1)
 #endif
-
+#define BUFFER_LENGTH 4096
 
 struct socket_info_struct socket_info;
 
@@ -114,7 +125,7 @@ static gint socket_fd_close			(gint sock);
 
 
 
-void send_open_command(gint sock, gint argc, gchar **argv)
+static void send_open_command(gint sock, gint argc, gchar **argv)
 {
 	gint i;
 	gchar *filename;
@@ -140,7 +151,10 @@ void send_open_command(gint sock, gint argc, gchar **argv)
 		g_free(col);
 	}
 
-	socket_fd_write_all(sock, "open\n", 5);
+	if (cl_options.readonly) /* append "ro" to denote readonly status for new docs */
+		socket_fd_write_all(sock, "openro\n", 7);
+	else
+		socket_fd_write_all(sock, "open\n", 5);
 
 	for (i = 1; i < argc && argv[i] != NULL; i++)
 	{
@@ -184,6 +198,53 @@ static void remove_socket_link_full(void)
 #endif
 
 
+static void socket_get_document_list(gint sock)
+{
+	gchar buf[BUFFER_LENGTH];
+	gint n_read;
+
+	if (sock < 0)
+		return;
+
+	socket_fd_write_all(sock, "doclist\n", 8);
+
+	do
+	{
+		n_read = socket_fd_read(sock, buf, BUFFER_LENGTH);
+		/* if we received ETX (end-of-text), there is nothing else to read, so cut that
+		 * byte not to output it and to be sure not to validate the loop condition */
+		if (n_read > 0 && buf[n_read - 1] == '\3')
+			n_read--;
+		if (n_read > 0)
+			fwrite(buf, 1, n_read, stdout);
+	}
+	while (n_read >= BUFFER_LENGTH);
+}
+
+
+#ifndef G_OS_WIN32
+static void check_socket_permissions(void)
+{
+	GStatBuf socket_stat;
+
+	if (g_lstat(socket_info.file_name, &socket_stat) == 0)
+	{	/* If the user id of the process is not the same as the owner of the socket
+		 * file, then ignore this socket and start a new session. */
+		if (socket_stat.st_uid != getuid())
+		{
+			const gchar *msg = _(
+	/* TODO maybe this message needs a rewording */
+	"Geany tried to access the Unix Domain socket of another instance running as another user.\n"
+	"This is a fatal error and Geany will now quit.");
+			g_warning("%s", msg);
+			dialogs_show_msgbox(GTK_MESSAGE_ERROR, "%s", msg);
+			exit(1);
+		}
+	}
+}
+#endif
+
+
 /* (Unix domain) socket support to replace the old FIFO code
  * (taken from Sylpheed, thanks)
  * Returns the created socket, -1 if an error occurred or -2 if another socket exists and files
@@ -191,16 +252,11 @@ static void remove_socket_link_full(void)
 gint socket_init(gint argc, gchar **argv)
 {
 	gint sock;
-
 #ifdef G_OS_WIN32
 	HANDLE hmutex;
-	/* we need a mutex name which is unique for the used configuration dir,
-	 * but we can't use the whole path, so build a hash on the configdir and use it */
-	gchar *mutex_name = g_strdup_printf("Geany%d", g_str_hash(app->configdir));
-
+	HWND hwnd;
 	socket_init_win32();
-	hmutex = CreateMutexA(NULL, FALSE, mutex_name);
-	g_free(mutex_name);
+	hmutex = CreateMutexA(NULL, FALSE, "Geany");
 	if (! hmutex)
 	{
 		geany_debug("cannot create Mutex\n");
@@ -208,6 +264,12 @@ gint socket_init(gint argc, gchar **argv)
 	}
 	if (GetLastError() != ERROR_ALREADY_EXISTS)
 	{
+		/* To support multiple instances with different configuration directories (as we do on
+		 * non-Windows systems) we would need to use different port number s but it might be
+		 * difficult to get a port number which is unique for a configuration directory (path)
+		 * and which is unused. This port number has to be guessed by the first and new instance
+		 * and the only data is the configuration directory path.
+		 * For now we use one port number, that is we support only one instance at all. */
 		sock = socket_fd_open_inet(REMOTE_CMD_PORT);
 		if (sock < 0)
 			return 0;
@@ -218,25 +280,34 @@ gint socket_init(gint argc, gchar **argv)
 	if (sock < 0)
 		return -1;
 #else
-	gchar *display_name = gdk_get_display();
-	gchar *hostname = utils_get_hostname();
+	gchar *display_name = NULL;
+	const gchar *hostname = g_get_host_name();
+	GdkDisplay *display = gdk_display_get_default();
 	gchar *p;
 
+	if (display != NULL)
+		display_name = g_strdup(gdk_display_get_name(display));
 	if (display_name == NULL)
 		display_name = g_strdup("NODISPLAY");
 
 	/* these lines are taken from dcopc.c in kdelibs */
 	if ((p = strrchr(display_name, '.')) > strrchr(display_name, ':') && p != NULL)
 		*p = '\0';
-	while ((p = strchr(display_name, ':')) != NULL)
-		*p = '_';
+	/* remove characters that may not be acceptable in a filename */
+	for (p = display_name; *p; p++)
+	{
+		if (*p == ':' || *p == '/')
+			*p = '_';
+	}
 
 	if (socket_info.file_name == NULL)
 		socket_info.file_name = g_strdup_printf("%s%cgeany_socket_%s_%s",
 			app->configdir, G_DIR_SEPARATOR, hostname, display_name);
 
 	g_free(display_name);
-	g_free(hostname);
+
+	/* check whether the real user id is the same as this of the socket file */
+	check_socket_permissions();
 
 	sock = socket_fd_connect_unix(socket_info.file_name);
 	if (sock < 0)
@@ -247,11 +318,24 @@ gint socket_init(gint argc, gchar **argv)
 #endif
 
 	/* remote command mode, here we have another running instance and want to use it */
+
+#ifdef G_OS_WIN32
+	/* first we send a request to retrieve the window handle and focus the window */
+	socket_fd_write_all(sock, "window\n", 7);
+	if (socket_fd_read(sock, (gchar *)&hwnd, sizeof(hwnd)) == sizeof(hwnd))
+		SetForegroundWindow(hwnd);
+#endif
+	/* now we send the command line args */
 	if (argc > 1)
 	{
 		send_open_command(sock, argc, argv);
 	}
 
+	if (cl_options.list_documents)
+	{
+		socket_get_document_list(sock);
+	}
+
 	socket_fd_close(sock);
 	return -2;
 }
@@ -259,7 +343,8 @@ gint socket_init(gint argc, gchar **argv)
 
 gint socket_finalize(void)
 {
-	if (socket_info.lock_socket < 0) return -1;
+	if (socket_info.lock_socket < 0)
+		return -1;
 
 	if (socket_info.lock_socket_tag > 0)
 		g_source_remove(socket_info.lock_socket_tag);
@@ -341,13 +426,13 @@ static gint socket_fd_open_unix(const gchar *path)
 	real_path = g_strdup_printf("%s%cgeany_socket.%08x",
 		g_get_tmp_dir(), G_DIR_SEPARATOR, g_random_int());
 
-	if (utils_is_file_writeable(real_path) != 0)
-	{	/* if real_path is not writable for us, fall back to ~/.geany/geany_socket_*_* */
+	if (utils_is_file_writable(real_path) != 0)
+	{	/* if real_path is not writable for us, fall back to ~/.config/geany/geany_socket_*_* */
 		/* instead of creating a symlink and print a warning */
 		g_warning("Socket %s could not be written, using %s as fallback.", real_path, path);
-		setptr(real_path, g_strdup(path));
+		SETPTR(real_path, g_strdup(path));
 	}
-	/* create a symlink in e.g. ~/.geany/geany_socket_hostname__0 to /tmp/geany_socket.499602d2 */
+	/* create a symlink in e.g. ~/.config/geany/geany_socket_hostname__0 to /tmp/geany_socket.499602d2 */
 	else if (symlink(real_path, path) != 0)
 	{
 		perror("symlink");
@@ -359,8 +444,6 @@ static gint socket_fd_open_unix(const gchar *path)
 	addr.sun_family = AF_UNIX;
 	strncpy(addr.sun_path, real_path, sizeof(addr.sun_path) - 1);
 
-	g_free(real_path);
-
 	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)
 	{
 		perror("bind");
@@ -375,6 +458,10 @@ static gint socket_fd_open_unix(const gchar *path)
 		return -1;
 	}
 
+	g_chmod(real_path, 0600);
+
+	g_free(real_path);
+
 	return sock;
 }
 #endif
@@ -392,12 +479,12 @@ static gint socket_fd_close(gint fd)
 #ifdef G_OS_WIN32
 static gint socket_fd_open_inet(gushort port)
 {
-	SockDesc sock;
+	SOCKET sock;
 	struct sockaddr_in addr;
 	gchar val;
 
 	sock = socket(AF_INET, SOCK_STREAM, 0);
-	if (! SOCKET_IS_VALID(sock))
+	if (G_UNLIKELY(! SOCKET_IS_VALID(sock)))
 	{
 		geany_debug("fd_open_inet(): socket() failed: %d\n", WSAGetLastError());
 		return -1;
@@ -436,11 +523,11 @@ static gint socket_fd_open_inet(gushort port)
 
 static gint socket_fd_connect_inet(gushort port)
 {
-	SockDesc sock;
+	SOCKET sock;
 	struct sockaddr_in addr;
 
 	sock = socket(AF_INET, SOCK_STREAM, 0);
-	if (! SOCKET_IS_VALID(sock))
+	if (G_UNLIKELY(! SOCKET_IS_VALID(sock)))
 	{
 		geany_debug("fd_connect_inet(): socket() failed: %d\n", WSAGetLastError());
 		return -1;
@@ -484,21 +571,45 @@ static void handle_input_filename(const gchar *buf)
 		utf8_filename = g_strdup(buf);
 
 	locale_filename = utils_get_locale_from_utf8(utf8_filename);
-	g_free(utf8_filename);
 	if (locale_filename)
-		main_handle_filename(locale_filename);
+	{
+		if (g_str_has_suffix(locale_filename, ".geany"))
+		{
+			if (project_ask_close())
+				main_load_project_from_command_line(locale_filename, TRUE);
+		}
+		else
+			main_handle_filename(locale_filename);
+	}
+	g_free(utf8_filename);
 	g_free(locale_filename);
 }
 
 
+static gchar *build_document_list(void)
+{
+	GString *doc_list = g_string_new(NULL);
+	guint i;
+	const gchar *filename;
+
+	foreach_document(i)
+	{
+		filename = DOC_FILENAME(documents[i]);
+		g_string_append(doc_list, filename);
+		g_string_append_c(doc_list, '\n');
+	}
+	return g_string_free(doc_list, FALSE);
+}
+
+
 gboolean socket_lock_input_cb(GIOChannel *source, GIOCondition condition, gpointer data)
 {
 	gint fd, sock;
-	gchar buf[4096];
+	gchar buf[BUFFER_LENGTH];
 	struct sockaddr_in caddr;
-	guint caddr_len;
-
-	caddr_len = sizeof(caddr);
+	socklen_t caddr_len = sizeof(caddr);
+	GtkWidget *window = data;
+	gboolean popup = FALSE;
 
 	fd = g_io_channel_unix_get_fd(source);
 	sock = accept(fd, (struct sockaddr *)&caddr, &caddr_len);
@@ -508,19 +619,27 @@ gboolean socket_lock_input_cb(GIOChannel *source, GIOCondition condition, gpoint
 	{
 		if (strncmp(buf, "open", 4) == 0)
 		{
+			cl_options.readonly = strncmp(buf+4, "ro", 2) == 0; /* open in readonly? */
 			while (socket_fd_gets(sock, buf, sizeof(buf)) != -1 && *buf != '.')
 			{
-				handle_input_filename(g_strstrip(buf));
-			}
+				gsize buf_len = strlen(buf);
 
-#ifdef G_OS_WIN32
-			/* we need to bring the main window up with gtk_window_present() but this is not
-			 * enough, instead we need to iconify it so that gtk_window_deiconify() will
-			 * bring it in the foreground */
-			gtk_window_present(GTK_WINDOW(main_widgets.window));
-			gtk_window_iconify(GTK_WINDOW(main_widgets.window));
-#endif
-			gtk_window_deiconify(GTK_WINDOW(main_widgets.window));
+				/* remove trailing newline */
+				if (buf_len > 0 && buf[buf_len - 1] == '\n')
+					buf[buf_len - 1] = '\0';
+
+				handle_input_filename(buf);
+			}
+			popup = TRUE;
+		}
+		else if (strncmp(buf, "doclist", 7) == 0)
+		{
+			gchar *doc_list = build_document_list();
+			if (!EMPTY(doc_list))
+				socket_fd_write_all(sock, doc_list, strlen(doc_list));
+			/* send ETX (end-of-text) so reader knows to stop reading */
+			socket_fd_write_all(sock, "\3", 1);
+			g_free(doc_list);
 		}
 		else if (strncmp(buf, "line", 4) == 0)
 		{
@@ -540,7 +659,42 @@ gboolean socket_lock_input_cb(GIOChannel *source, GIOCondition condition, gpoint
 				cl_options.goto_column = atoi(buf);
 			}
 		}
+#ifdef G_OS_WIN32
+		else if (strncmp(buf, "window", 6) == 0)
+		{
+#	if GTK_CHECK_VERSION(3, 0, 0)
+			HWND hwnd = (HWND) gdk_win32_window_get_handle(gtk_widget_get_window(window));
+#	else
+			HWND hwnd = (HWND) gdk_win32_drawable_get_handle(
+				GDK_DRAWABLE(gtk_widget_get_window(window)));
+#	endif
+			socket_fd_write(sock, (gchar *)&hwnd, sizeof(hwnd));
+		}
+#endif
 	}
+
+	if (popup)
+	{
+#ifdef GDK_WINDOWING_X11
+		GdkWindow *x11_window = gtk_widget_get_window(window);
+
+		/* Set the proper interaction time on the window. This seems necessary to make
+		 * gtk_window_present() really bring the main window into the foreground on some
+		 * window managers like Gnome's metacity.
+		 * Code taken from Gedit. */
+#	if GTK_CHECK_VERSION(3, 0, 0)
+		if (GDK_IS_X11_WINDOW(x11_window))
+#	endif
+		{
+			gdk_x11_window_set_user_time(x11_window, gdk_x11_get_server_time(x11_window));
+		}
+#endif
+		gtk_window_present(GTK_WINDOW(window));
+#ifdef G_OS_WIN32
+		gdk_window_show(gtk_widget_get_window(window));
+#endif
+	}
+
 	socket_fd_close(sock);
 
 	return TRUE;
@@ -606,20 +760,22 @@ static gint socket_fd_check_io(gint fd, GIOCondition cond)
 
 #ifdef G_OS_UNIX
 	/* checking for non-blocking mode */
-
 	flags = fcntl(fd, F_GETFL, 0);
 	if (flags < 0)
 	{
 		perror("fcntl");
 		return 0;
 	}
-
 	if ((flags & O_NONBLOCK) != 0)
 		return 0;
 #endif
 
 	FD_ZERO(&fds);
+#ifdef G_OS_WIN32
+	FD_SET((SOCKET)fd, &fds);
+#else
 	FD_SET(fd, &fds);
+#endif
 
 	if (cond == G_IO_IN)
 	{
@@ -636,7 +792,7 @@ static gint socket_fd_check_io(gint fd, GIOCondition cond)
 	}
 	else
 	{
-		geany_debug("Socket IO timeout\n");
+		geany_debug("Socket IO timeout");
 		return -1;
 	}
 }
@@ -674,5 +830,3 @@ gint socket_fd_write(gint fd, const gchar *buf, gint len)
 
 
 #endif
-
-
